/*
 * Copyright (C) Sergey P. Derevyago, 2005.
 *
 * Permission to copy, use, modify, sell and distribute this software is granted
 * provided this copyright notice appears in all copies.
 * This software is provided "as is" without express or implied warranty, and
 * with no claim as to its suitability for any purpose.
 *
 */

package ders;

import java.io.InputStream;
import java.io.OutputStream;
import java.math.BigInteger;
import java.util.Arrays;

/**
 * A production quality implementation of DersCrypt v1.1 symmetric-key block
 * cipher.
 *
 * @see <a href="http://ders.stml.net/crypt/derscrypt/derscrypt.html">Crypto
 * algorithm DersCrypt</a> -- DersCrypt home page.
 * @version 1.1 2005-12-24
 * @author Sergey P. Derevyago
 */
public final class Crypt
{
 /**
  * Encrypts <code>in</code> and writes the result to <code>out</code>.
  *
  * @param key key to use
  * @param in stream to read plain text from
  * @param out stream to write encrypted text to
  * @return number of bytes read from <code>in</code>
  * @throws ders.Exception in case of errors
  */
 public static long streamEncrypt(Key key, InputStream in, OutputStream out)
   throws Exception
 {
  try {
      long readCnt=0;
      int minLen=minBlockLength(key.getBytes().length);

      byte[] buf1=readMax(in, minLen);
      readCnt+=buf1.length;
      
      if (buf1.length<minLen) throw new InsufficientDataException(minLen); 

      for (byte[] prevEncr=null; buf1.length>0; ) {
          byte[] plain, buf2=readMax(in, minLen);
          readCnt+=buf2.length;

          if (buf2.length==minLen) {
             plain=buf1;
             buf1=buf2;
          }
          else {
               plain=new byte[buf1.length+buf2.length];
               System.arraycopy(buf1, 0, plain, 0, buf1.length);
               System.arraycopy(buf2, 0, plain, buf1.length, buf2.length);

               buf1=new byte[0];
          }

          byte[] encr=blockEncrypt(key, plain, prevEncr);
          prevEncr=encr;

          writeInt16(out, encr.length);
          out.write(encr);
      }

      return readCnt;
  }
  catch (Exception e) { throw e; }
  catch (Throwable t) { throw new Exception(t); }
 }

 /**
  * Decrypts <code>in</code> and writes the result to <code>out</code>.
  *
  * @param key key to use
  * @param in stream to read encrypted text from
  * @param out stream to write decrypted text to
  * @return number of bytes read from <code>in</code>
  * @throws ders.Exception in case of errors
  */
 public static long streamDecrypt(Key key, InputStream in, OutputStream out)
   throws Exception
 {
  try {
      long readCnt=0;
      int minLen=minBlockLength(key.getBytes().length);

      byte[] prevEncr=null;
      for (int len; (len=readInt16(in))!=-1; ) {
          readCnt+=2;
          if (len<=minLen || len>=minLen*3)
             throw new InvalidBlockFormatDataException();

          byte[] encr=readMax(in, len);
          readCnt+=encr.length;

          if (encr.length!=len) throw new InvalidBlockFormatDataException();

          byte[] decr=blockDecrypt(key, encr, prevEncr);
          prevEncr=encr;

          out.write(decr);
      }

      return readCnt;
  }
  catch (Exception e) { throw e; }
  catch (Throwable t) { throw new Exception(t); }
 }

 /**
  * Encrypts a block of plain text.
  *
  * @param key key to use
  * @param plain block of plain text to encrypt
  * @return encrypted block
  * @throws ders.Exception in case of errors
  */
 public static byte[] blockEncrypt(Key key, byte[] plain) throws Exception
 {
  try {
      int keyLength=key.getBytes().length;

      int minLen=minBlockLength(keyLength);
      if (plain.length<minLen)
         throw new InsufficientDataException(minLen);

      int maxLen=maxBlockLength(keyLength);
      if (plain.length>maxLen)
         throw new ExcessiveDataException(maxLen);

      return blockEncrypt(key, Util.byteArrayCopy(plain), null);
  }
  catch (Exception e) { throw e; }
  catch (Throwable t) { throw new Exception(t); }
 }

 /**
  * Decrypts a block of encrypted text.
  *
  * @param key key to use
  * @param encr block of encrypted text to decrypt
  * @return decrypted block
  * @throws ders.Exception in case of errors
  */
 public static byte[] blockDecrypt(Key key, byte[] encr) throws Exception
 {
  try {
      int minLen=minBlockLength(key.getBytes().length)+1;
      if (encr.length<minLen)
         throw new InsufficientDataException(minLen);
         
      int maxLen=minLen*3-1;
      if (encr.length>maxLen)
         throw new ExcessiveDataException(maxLen);

      return blockDecrypt(key, Util.byteArrayCopy(encr), null);
  }
  catch (Exception e) { throw e; }
  catch (Throwable t) { throw new Exception(t); }
 }

 /**
  * Checks whether passed bytes can be used as key.
  *
  * @param key bytes to check
  * @throws InvalidKeyException if the key is not valid
  */
 public static void checkKey(byte[] key) throws InvalidKeyException
 {
  if (key[0]==0) throw new LeadingZeroKeyException();

  if ((key[key.length-1]&1)==0) throw new EvenValueKeyException();
 }

 /**
  * Returns minimal data block length required for passed key length.
  *
  * @param keyLength key length in bytes
  * @return minimal data block length
  */
 public static int minBlockLength(int keyLength)
 {
  if (keyLength<16) keyLength=16;       // 128->40
  else if (keyLength>64) keyLength=64;  // 512->100

  int n=40+((keyLength-16)*5+2)/4;
  return keyLength*n;
 }

 /**
  * Returns maximum data block length that can be encrypted by blockEncrypt()
  * corresponding to passed key length.
  *
  * @param keyLength key length in bytes
  * @return maximum data block length
  */
 public static int maxBlockLength(int keyLength)
 {
  return minBlockLength(keyLength)*2-1;
 }
 
 private static final int[] BIT_MASKS={
   0x80, 0x40, 0x20, 0x10, 0x08, 0x04, 0x02, 0x01
 };

 private static byte[] blockEncrypt(Key key, byte[] plain, byte[] prevEncr)
   throws Exception
 {
  byte[] hash=computeHash(key, plain, prevEncr);

  BigInteger aPr1=new BigInteger(1, appendHash(plain, hash));

  boolean[] substOne=new boolean[1];
  BigInteger aPr2=basicEncrypt(key, aPr1, substOne);
  byte[] encr=stripLeadingZeros(aPr2.toByteArray());

  if (substOne[0] && blockDecryptImpl(key, encr, prevEncr, false)!=null)
     throw new CannotEncryptException();

  return encr;
 }

 private static byte[] blockDecrypt(Key key, byte[] encr, byte[] prevEncr)
   throws Exception
 {
  byte[] decr=blockDecryptImpl(key, encr, prevEncr, false);
  if (decr!=null) return decr;

  decr=blockDecryptImpl(key, encr, prevEncr, true);
  if (decr==null) throw new IncorrectKeyException();

  return decr;
 }
 
 private static byte[] blockDecryptImpl(Key key, byte[] encr, byte[] prevEncr,
   boolean substOne)
 {
  BigInteger aPr2=new BigInteger(1, encr);
  BigInteger aPr1=basicDecrypt(key, aPr2, substOne);
  if (aPr1==null) return null;

  byte[] decr=stripLeadingZeros(aPr1.toByteArray());

  byte[][] ph=extractHash(decr, key.getBytes().length);
  byte[] plain=ph[0], hash=ph[1];

  byte[] hash2=computeHash(key, plain, prevEncr);
  if (!Arrays.equals(hash, hash2)) return null;

  return plain;
 }
 
 private static BigInteger basicEncrypt(Key key, BigInteger a1,
   boolean[] substOne)
 {
  BigInteger[] rems=key.computeRemainders(a1);
  BigInteger[] rems2=permute(rems, key, false);

  if (rems2[0].equals(BigInteger.ZERO)) {
     rems2[0]=BigInteger.ONE;
     substOne[0]=true;
  }
  else substOne[0]=false;

  BigInteger a2=computeNumber(rems2, key.getValue());
  return a2;
 }

 private static BigInteger basicDecrypt(Key key, BigInteger a2,
   boolean substOne)
 {
  BigInteger[] rems=key.computeRemainders(a2);

  if (substOne) {
     if (rems[0].equals(BigInteger.ONE)) rems[0]=BigInteger.ZERO;
     else return null;
  }

  BigInteger[] rems2=permute(rems, key, true);
  BigInteger a1=computeNumber(rems2, key.getValue());

  return a1;
 }

 private static BigInteger computeNumber(BigInteger[] rems, BigInteger b)
 {
  BigInteger a=rems[0];

  for (int i=1; i<rems.length; i++) {
      a=a.multiply(b);
      a=a.add(rems[i]);
  }

  return a;
 }
 
 private static BigInteger[] permute(BigInteger[] rems, Key key,
   boolean inverse)
 {
  int len=rems.length;
  BigInteger[] ret=new BigInteger[len];

  int perm[]=key.getPermutation(len);
  if (inverse) for (int i=0; i<len; i++) ret[i]=rems[perm[i]];
  else for (int i=0; i<len; i++) ret[perm[i]]=rems[i];

  return ret;
 }
 
 private static byte[] permuteBits(byte[] bits, Key key)
 {
  byte[] ret=new byte[bits.length];

  int perm[]=key.getPermutation(bits.length*8);

  for (int i=0, j=0; i<bits.length; i++) {
      for (int mask=0x80; mask!=0; mask>>=1) {
          int n=perm[j++];
          if ((bits[i]&mask)!=0)
             ret[n/8]|=BIT_MASKS[n%8];
      }
  }

  return ret;
 }

 private static byte[] computeHash(Key key, byte[] data, byte[] prevData)
 {
  int len=key.getBytes().length;

  byte[] bits;
  if (prevData!=null) bits=extractBits(prevData, len);
  else {
       bits=new byte[len];
       for (int i=0; i<len; i++) bits[i]=0x55;
  }

  BigInteger a1=new BigInteger(1, appendHash(data, permuteBits(bits, key)));
  BigInteger a2=basicEncrypt(key, a1, new boolean[1]);
  byte[] encr=stripLeadingZeros(a2.toByteArray());

  byte[] hash=permuteBits(extractBits(encr, len), key);
  if (hash[0]==0) hash[0]=1;

  return hash;
 }
 
 private static byte[] extractBits(byte[] data, int bitsLen)
 {
  byte[] ret=new byte[bitsLen];

  for (int i=0, j=1; i<bitsLen; i++) {
      for (int mask=0x80; mask!=0; mask>>=1) {
          int n=(j++)*(data.length*8-1)/(bitsLen*8+1);
          if ((data[n/8]&BIT_MASKS[n%8])!=0)
             ret[i]|=mask;
      }
  }

  return ret;
 }

 private static byte[] appendHash(byte[] data, byte[] hash)
 {
  byte[] ret=new byte[data.length+hash.length];

  int len=hash.length, mid=len/2;
  System.arraycopy(hash, 0, ret, 0, len-mid);
  System.arraycopy(data, 0, ret, len-mid, data.length);
  System.arraycopy(hash, len-mid, ret, ret.length-mid, mid);

  return ret;
 }
 
 private static byte[][] extractHash(byte[] bytes, int hashLen)
 {
  byte[] data=new byte[bytes.length-hashLen];
  byte[] hash=new byte[hashLen];

  int mid=hashLen/2;
  System.arraycopy(bytes, 0, hash, 0, hashLen-mid);
  System.arraycopy(bytes, hashLen-mid, data, 0, data.length);
  System.arraycopy(bytes, bytes.length-mid, hash, hashLen-mid, mid);

  return new byte[][]{data, hash};
 }
 
 private static byte[] stripLeadingZeros(byte[] bytes)
 {
  if (bytes[0]!=0) return bytes;

  int i;
  for (i=1; i<bytes.length && bytes[i]==0; i++)
      ;

  return Util.byteSubArrayCopy(bytes, i, bytes.length-i);
 }

 private static byte[] readMax(InputStream in, int maxLen)
   throws java.lang.Exception
 {
  byte[] buf=new byte[maxLen];
  int cnt=in.read(buf);

  if (cnt==buf.length) return buf;
  if (cnt==-1) return new byte[0];
  return Util.byteSubArrayCopy(buf, 0, cnt);
 }

 private static void writeInt16(OutputStream out, int val)
   throws java.lang.Exception
 {
  out.write((val>>>8)&0xFF);
  out.write(val&0xFF);
 }

 private static int readInt16(InputStream in) throws java.lang.Exception
 {
  int ch1=in.read();
  if (ch1==-1) return -1;

  int ch2=in.read();
  if (ch2==-1) throw new InvalidBlockFormatDataException();
   
  return (ch1<<8)+ch2;
 }
}

