/*
 * Copyright (C) Sergey P. Derevyago, 2005.
 *
 * Permission to copy, use, modify, sell and distribute this software is granted
 * provided this copyright notice appears in all copies.
 * This software is provided "as is" without express or implied warranty, and
 * with no claim as to its suitability for any purpose.
 *
 */

package ders;

import java.math.BigInteger;
import java.util.LinkedList;

/**
 * Holds key for DersCrypt cipher and provides key-dependant functionality.
 */
public final class Key
{
 /**
  * Creates key from passed bytes.
  *
  * @param key bytes to create key from
  * @throws InvalidKeyException if the bytes cannot be used as key
  */
 public Key(byte[] key) throws InvalidKeyException
 {
  try {
      bytes=Util.byteArrayCopy(key);
      Crypt.checkKey(bytes);

      powers[0]=new BigInteger(1, bytes);

      for (int i=1; i<powers.length; i++)
          powers[i]=powers[i-1].multiply(powers[i-1]);
  }
  catch (InvalidKeyException ike) { throw ike; }
  catch (Throwable t) { throw new InvalidKeyException(t); }
 }

 /**
  * Returns key bytes.
  *
  * @return copy of key bytes
  */
 public byte[] getBytes()
 {
  return Util.byteArrayCopy(bytes);
 }

 /**
  * Returns key value.
  *
  * @return key value
  */
 public BigInteger getValue()
 {
  return powers[0];
 }

 /**
  * Splits <code>a</code> into a sequence of remainders of division by the key
  * value.
  *
  * @param a number to split
  * @return array of remainders
  */
 public BigInteger[] computeRemainders(BigInteger a)
 {
  int depth;
  for (depth=-1; a.compareTo(powers[depth+1])>=0; )
      depth++;

  LinkedList rems=new LinkedList();
  computeRemainders(a, depth, rems);

  return (BigInteger[])rems.toArray(new BigInteger[rems.size()]);
 }

 /**
  * Returns permutation of passed length created from the key value.
  *
  * @param length length of permutation
  * @return array of positions
  */
 public int[] getPermutation(int length)
 {
  int[] perm=cache.get(length);

  if (perm==null) {
     perm=createPermutation(length, getValue());
     cache.put(perm);
  }

  return Util.intArrayCopy(perm);
 }

 private static class Cache
 {
  public int[] get(int len)
  {
   for (int i=0; i<size; i++)
       if (elems[i].length==len) return elems[i];

   return null;
  }

  public void put(int[] elem)
  {
   if (size>=elems.length) {
      int[][] newEls=new int[size*2][];
      
      for (int i=0; i<size; i++) newEls[i]=elems[i];
      elems=newEls;
   }

   elems[size++]=elem;
  }

  private int[][] elems=new int[10][];
  private int size=0;
 }

 private byte[] bytes;
 private BigInteger[] powers=new BigInteger[9];  // last is value^256
 private Cache cache=new Cache();

 private static int[] createPermutation(int len, BigInteger b)
 {
  BigInteger b2=b;
  int[] ret=new int[len], val=new int[len];

  int med=len-len/2;
  for (int i=0; i<len; i++) {
      ret[i]=-2;

      int pos= (i<med) ? len-med+i : i-med;
      val[pos]=i;
  }

  int[] points={0, len-med, len};

  for (int p=0; p<points.length-1; p++) {
      int pbeg=points[p], plen=points[p+1]-pbeg;

     loop:
      for (int i=0, sub=0; i<plen; i++) {
          if (i==plen-2 && sub==0) {  // don't left last element alone
             for (int j=0; ; j++) {
                 if (ret[pbeg+j]==-2) {
                    if (ret[pbeg+j+1]==-2) {
                       ret[pbeg+j]=val[pbeg+i+1];
                       ret[pbeg+j+1]=val[pbeg+i];
                       break loop;
                    }

                    break;
                 }
             }
          }

          BigInteger div=BigInteger.valueOf(plen-i-sub);
          if (b2.compareTo(div)<0) b2=b;  // restart from b

          BigInteger[] qr=b2.divideAndRemainder(div);
          b2=qr[0];
          int rem=qr[1].intValue();

          for (int j=0, k=-1, v=val[pbeg+i]; ; j++) {
              if (ret[pbeg+j]==-2 && !(j>0 && v==ret[pbeg+j-1]+1) && ++k==rem) {
                 ret[pbeg+j]=v;
                 sub= (j+1<plen && ret[pbeg+j+1]==-2) ? 1 : 0;

                 break;
              }
          }
      }
  }

  return ret;
 }

 private void computeRemainders(BigInteger a, int depth, LinkedList rems)
 {
  if (depth<0) {
     rems.addLast(a);
     return;
  }

  BigInteger[] qr=a.divideAndRemainder(powers[depth]);
  
  if (depth==0) {
     if (!qr[0].equals(BigInteger.ZERO) || rems.size()>0) rems.addLast(qr[0]);
     if (!qr[1].equals(BigInteger.ZERO) || rems.size()>0) rems.addLast(qr[1]);

     return;
  }

  computeRemainders(qr[0], depth-1, rems);
  computeRemainders(qr[1], depth-1, rems);
 }
}

