/*
 * Copyright (C) Sergey P. Derevyago, 2005.
 *
 * Permission to copy, use, modify, sell and distribute this software is granted
 * provided this copyright notice appears in all copies.
 * This software is provided "as is" without express or implied warranty, and
 * with no claim as to its suitability for any purpose.
 *
 */

/** @file
 * Implements multiple precision unsigned integer arithmetic via GMP library.
 *
 * GMP home page: http://www.swox.com/gmp/
 */

#ifndef __UINT_HPP__
#define __UINT_HPP__

#include <stdio.h>
#include <gmp.h>
#include "sh_ptr.hpp"
#include "util.hpp"

class UInt {
      mpz_t rep;

      void init() { mpz_init(rep); }

 public:
      static const UInt zero;
      static const UInt one;

      UInt() { init(); }
      UInt(uint val) { mpz_init_set_ui(rep, val); }
      UInt(const void* ptr, int size);
      UInt(const UInt& r) { mpz_init_set(rep, r.rep); }
      ~UInt() { mpz_clear(rep); }

      void* operator new(size_t) { return fixed_alloc<UInt>::alloc(); }
      void operator delete(void* ptr, size_t) { fixed_alloc<UInt>::free(ptr); }

      UInt& operator=(const UInt& r);
      UInt& operator+=(const UInt& r);
      UInt& operator-=(const UInt& r);
      UInt& operator*=(const UInt& r);
      UInt& operator/=(const UInt& r);
      UInt& operator%=(const UInt& r);

      bool operator==(const UInt& r) const { return mpz_cmp(rep, r.rep)==0; }
      bool operator!=(const UInt& r) const { return mpz_cmp(rep, r.rep)!=0; }
      bool operator<=(const UInt& r) const { return mpz_cmp(rep, r.rep)<=0; }
      bool operator>=(const UInt& r) const { return mpz_cmp(rep, r.rep)>=0; }
      bool operator<(const UInt& r) const { return mpz_cmp(rep, r.rep)<0; }
      bool operator>(const UInt& r) const { return mpz_cmp(rep, r.rep)>0; }

      int getSize() const { return (mpz_sizeinbase(rep, 16)+1)/2; }

      void fromUInt(uint val) { mpz_set_ui(rep, val); }
      uint toUInt() const { return mpz_get_ui(rep); }
      void fromBytes(const void* ptr, int size);
      int toBytes(void* ptr) const;
      void fromStr(char* str, int base=10) { mpz_set_str(rep, str, base); }
      void toStr(char* str, int base=10) const { mpz_get_str(str, base, rep); }

      static void qr(const UInt& a, const UInt& b, UInt& q, UInt& r);

      static const char* cpuString();
};

typedef sh_ptr<UInt> ShUInt;

inline ShUInt newUInt() { return ShUInt(new UInt()); }
inline ShUInt newUInt(uint val) { return ShUInt(new UInt(val)); }
inline ShUInt newUInt(const void* ptr, int size);
inline ShUInt newUInt(const UInt& r) { return ShUInt(new UInt(r)); }

ShUInt operator+(const ShUInt& lo, const ShUInt& ro);
ShUInt operator-(const ShUInt& lo, const ShUInt& ro);
ShUInt operator*(const ShUInt& lo, const ShUInt& ro);
ShUInt operator/(const ShUInt& lo, const ShUInt& ro);
ShUInt operator%(const ShUInt& lo, const ShUInt& ro);

inline bool operator==(const ShUInt& lo, const ShUInt& ro) { return *lo==*ro; }
inline bool operator!=(const ShUInt& lo, const ShUInt& ro) { return *lo!=*ro; }
inline bool operator<=(const ShUInt& lo, const ShUInt& ro) { return *lo<=*ro; }
inline bool operator>=(const ShUInt& lo, const ShUInt& ro) { return *lo>=*ro; }
inline bool operator<(const ShUInt& lo, const ShUInt& ro) { return *lo<*ro; }
inline bool operator>(const ShUInt& lo, const ShUInt& ro) { return *lo>*ro; }

inline UInt::UInt(const void* ptr, int size)
{
 init();
 fromBytes(ptr, size);
}

inline UInt& UInt::operator=(const UInt& r)
{
 mpz_set(rep, r.rep);
 return *this;
}

inline UInt& UInt::operator+=(const UInt& r)
{
 mpz_add(rep, rep, r.rep);
 return *this;
}

inline UInt& UInt::operator-=(const UInt& r)
{
 mpz_sub(rep, rep, r.rep);
 return *this;
}

inline UInt& UInt::operator*=(const UInt& r)
{
 mpz_mul(rep, rep, r.rep);
 return *this;
}

inline UInt& UInt::operator/=(const UInt& r)
{
 mpz_tdiv_q(rep, rep, r.rep);
 return *this;
}

inline UInt& UInt::operator%=(const UInt& r)
{
 mpz_tdiv_r(rep, rep, r.rep);
 return *this;
}

inline void UInt::fromBytes(const void* ptr, int size)
{
 mpz_import(rep, size, 1, 1, 1, 0, ptr);
}

inline void UInt::qr(const UInt& a, const UInt& b, UInt& q, UInt& r)
{
 mpz_tdiv_qr(q.rep, r.rep, a.rep, b.rep);
}

inline ShUInt newUInt(const void* ptr, int size)
{
 return ShUInt(new UInt(ptr, size));
}

#endif

